@@ -1,5 +1,8 @@
 /* atmdump.c - ATM raw cell dumper */
 
+/* Written 1995 by Werner Almesberger, EPFL-LRC */
+
+
 #include <stdlib.h>
 #include <stdio.h>
 #include <unistd.h>
@@ -72,6 +75,8 @@ int main(int argc,char **argv)
 	    case 'c':
 		clp = 1;
 		break;
+	    default:
+		usage(name);
 	}
     if (argc != optind+3) usage(name);
     if ((s = socket(PF_ATMPVC,SOCK_DGRAM,ATM_AAL0)) < 0) {
