@@ -13,11 +13,9 @@
 #include <linux/atm.h>
 #include <linux/atmsvc.h>
 
+#include "atmd.h"
 #include "q2931.h"
 
-#include "diag.h"
-#include "common.h"
-#include "timer.h"
 #include "proto.h"
 #include "io.h"
 
@@ -175,7 +173,7 @@ void to_net(void *msg,int size)
       (((unsigned char *) msg)[2] << 16) |
       (((unsigned char *) msg)[3] << 8) | ((unsigned char *) msg)[4],
       size);
-    if (dump) {
+    if (debug) {
 	for (i = 0; i < size; i++) printf("%02X ",((unsigned char *) msg)[i]);
 	if (size) putchar('\n');
     }
@@ -232,7 +230,9 @@ void poll_loop(void)
     while (1) {
 	set = perm;
 	ret = select(fds,&set,NULL,NULL,next_timer());
-	if (ret < 0 && errno != EINTR) perror("select");
+	if (ret < 0) {
+	    if (errno != EINTR) perror("select");
+	}
 	else {
 	    diag(COMPONENT,DIAG_DEBUG,"----------");
 	    gettimeofday(&now,NULL);
