@@ -10,18 +10,16 @@
 #include <linux/atmsvc.h>
 #include <linux/atmdev.h>
 
+#include "atmd.h"
 #include "q2931.h"
 #include "qlib.h"
 
-#include "diag.h"
-#include "common.h"
 #include "proto.h"
 #include "io.h"
-#include "timer.h"
 #include "timeout.h"
 
 
-#define COMPONENT "KSP"
+#define COMPONENT "KERNEL"
 
 
 extern int vci;
@@ -30,6 +28,8 @@ extern int vci;
 static int send_setup(SOCKET *sock)
 {
     static unsigned long call_ref = 0;
+    struct atm_bhli *bhli;
+    struct atm_blli *blli;
     SOCKET *walk;
     Q_DSC dsc;
     int size;
@@ -45,16 +45,101 @@ static int send_setup(SOCKET *sock)
     q_assign(&dsc,QF_call_ref,call_ref);
     q_assign(&dsc,QF_msg_type,QMSG_SETUP);
     q_assign(&dsc,QF_aal_type,5); /* AAL 5 */
-    q_assign(&dsc,QF_fw_pcr_01,1000); /* Fore doesn't want ATM_OC3_PCR */
-    q_assign(&dsc,QF_bw_pcr_01,1000); /* Fore doesn't want ATM_OC3_PCR */
+    if (sock->remote->sas_txtp.max_sdu)
+	q_assign(&dsc,QF_fw_max_sdu,sock->remote->sas_txtp.max_sdu);
+    if (sock->remote->sas_rxtp.max_sdu)
+	q_assign(&dsc,QF_bw_max_sdu,sock->remote->sas_rxtp.max_sdu);
+    q_assign(&dsc,QF_fw_pcr_01,1000); /* Fore doesn't like ATM_OC3_PCR */
+    q_assign(&dsc,QF_bw_pcr_01,1000); /* Fore doesn't like ATM_OC3_PCR */
     q_assign(&dsc,QF_best_effort,0);
     q_assign(&dsc,QF_bearer_class,3); /* BCOB-C */
     q_assign(&dsc,QF_upcc,0); /* p2p */
     q_assign(&dsc,QF_qos_fw,0); /* QOS 0 */
     q_assign(&dsc,QF_qos_bw,0);
-    /* send local address if available */
-    if (*sock->remote->sas_addr.prv)
-	q_write(&dsc,QF_cdpn_nsap,(void *) sock->remote->sas_addr.prv,
+    if (sock->local && *sock->local->sas_addr.prv) { /* improve @@@ */
+	q_assign(&dsc,QF_cgpn_plan,2);
+	q_assign(&dsc,QF_cgpn_type,0);
+	q_write(&dsc,QF_cgpn,(void *) sock->local->sas_addr.prv,ATM_ESA_LEN);
+    }
+    bhli = &sock->remote->sas_addr.bhli;
+    if (bhli->hl_type != ATM_HL_NONE) {
+	q_assign(&dsc,QF_hli_type,bhli->hl_type);
+	switch (bhli->hl_type) {
+	    case ATM_HL_ISO:
+		q_write(&dsc,QF_iso_hli,bhli->hl_info,bhli->hl_length);
+		break;
+	    case ATM_HL_USER:
+		q_write(&dsc,QF_iso_hli,bhli->hl_info,bhli->hl_length);
+		break;
+#ifdef UNI30
+	    case ATM_HL_HLP:
+		q_write(&dsc,QF_iso_hli,bhli->hl_info,4);
+		break;
+#endif
+	    case ATM_HL_VENDOR:
+		q_write(&dsc,QF_iso_hli,bhli->hl_info,7);
+		break;
+	    default:
+		diag(COMPONENT,DIAG_FATAL,"bad hl_type (%d)",bhli->hl_type);
+	}
+    }
+    blli = sock->remote->sas_addr.blli;
+    if (blli) { /* @@@ loop */
+	if (blli->l2_proto != ATM_L2_NONE) {
+	    q_assign(&dsc,QF_uil2_proto,blli->l2_proto);
+	    switch (blli->l2_proto) {
+		case ATM_L2_X25_LL:
+		case ATM_L2_X25_ML:
+		case ATM_L2_HDLC_ARM:
+		case ATM_L2_HDLC_NRM:
+		case ATM_L2_HDLC_ABM:
+		case ATM_L2_Q922:
+		case ATM_L2_ISO7776:
+		    if (blli->l2.itu.mode != ATM_IMD_NONE)
+			q_assign(&dsc,QF_l2_mode,blli->l2.itu.mode);
+		    if (blli->l2.itu.window)
+			q_assign(&dsc,QF_window_size,blli->l2.itu.window);
+		    break;
+		case ATM_L2_USER:
+		    q_assign(&dsc,QF_user_l2,blli->l2.user);
+		    break;
+		default:
+		    break;
+	    }
+	}
+	if (blli->l3_proto != ATM_L3_NONE) {
+	    q_assign(&dsc,QF_uil3_proto,blli->l3_proto);
+	    switch (blli->l3_proto) {
+		case ATM_L3_X25:
+		case ATM_L3_ISO8208:
+		case ATM_L3_X223:
+		    if (blli->l3.itu.mode != ATM_IMD_NONE)
+			q_assign(&dsc,QF_l3_mode,blli->l3.itu.mode);
+		    if (blli->l3.itu.def_size)
+			q_assign(&dsc,QF_def_pck_size,blli->l3.itu.def_size);
+		    if (blli->l3.itu.window)
+			q_assign(&dsc,QF_pck_win_size,blli->l3.itu.window);
+		    break;
+		case ATM_L3_TR9577:
+		    q_assign(&dsc,QF_ipi_high,blli->l3.tr9577.ipi >> 1);
+		    if ((blli->l3.tr9577.ipi >> 1) == 0x40)
+			q_assign(&dsc,QF_ipi_low,blli->l3.tr9577.ipi & 1);
+		    else if (!(blli->l3.tr9577.ipi & 1)) {
+			    q_write(&dsc,QF_oui,blli->l3.tr9577.snap,3);
+			    q_write(&dsc,QF_pid,blli->l3.tr9577.snap+3,2);
+			}
+		    break;
+		case ATM_L3_USER:
+		    q_assign(&dsc,QF_user_l3,blli->l3.user);
+		    break;
+		default:
+		    diag(COMPONENT,DIAG_FATAL,"bad l3_proto (%d)",
+		      blli->l3_proto);
+	    }
+	}
+    }
+    if (*sock->remote->sas_addr.prv) /* improve @@@ */
+	q_write(&dsc,QF_cdpn_esa,(void *) sock->remote->sas_addr.prv,
 	  ATM_ESA_LEN);
     if (net) {
 	q_assign(&dsc,QF_vpi,0);
@@ -91,7 +176,7 @@ static void send_release_complete(SOCKET *sock,unsigned char reason)
     q_create(&dsc,q_buffer,MAX_Q_MSG);
     q_assign(&dsc,QF_msg_type,QMSG_REL_COMP);
     q_assign(&dsc,QF_call_ref,sock->call_ref);
-    /* set cause */
+    if (reason) q_assign(&dsc,QF_cause,reason); /* @@@ more data */
     size = q_close(&dsc);
     to_signaling(q_buffer,size);
 }
@@ -162,18 +247,22 @@ static void dispatch(SOCKET *sock,struct atmsvc_msg *msg)
 	    new_state(next,ss_wait_rel);
 	    return;
 	case as_listen: /* NULL */
-	    if (sock) break;
-	    /* check address @@@ */
-	    error = 0;
-	    if (error) {
-		send_kernel(msg->vcc,0,as_close,error,NULL);
+	    {
+		struct sockaddr_atmsvc *sap;
+
+		if (sock) break;
+		sap = alloc_sap(&msg->local);
+		if (lookup_sap(sap)) {
+		    free_sap(sap);
+		    send_kernel(msg->vcc,0,as_close,-EADDRINUSE,NULL);
+		    return;
+		}
+		sock = new_sock(msg->vcc);
+		sock->local = alloc_sap(&msg->local);
+		send_kernel(sock->id,0,as_okay,0,NULL);
+		sock->state = ss_listening;
 		return;
 	    }
-	    sock = new_sock(msg->vcc);
-	    sock->local = alloc_sap(&msg->local);
-	    send_kernel(sock->id,0,as_okay,error,NULL);
-	    sock->state = ss_listening;
-	    return;
 	case as_close: /* all but ACCEPTING and WAIT_REL */
 	    if (sock && (sock->state == ss_accepting || sock->state ==
 	      ss_wait_rel)) break;
@@ -198,7 +287,13 @@ static void dispatch(SOCKET *sock,struct atmsvc_msg *msg)
 		    /* fall through */
 		case ss_indicated:
 		case ss_hold:
-		    send_release(sock,0); /* @@@ */
+		    send_release(sock,
+#ifdef UNI31
+		      ATM_CV_NORMAL_CLEAR
+#else
+		      ATM_CV_NORMAL_UNSPEC
+#endif
+		      );
 		    START_TIMER(sock,T308_1);
 		    new_state(sock,sock->state == ss_connected ? ss_rel_req :
 		      ss_wait_rel);
@@ -237,8 +332,11 @@ void from_kernel(struct atmsvc_msg *msg,int size)
       "in state %s",as_name[msg->type],(unsigned long) curr,msg->vcc,
       msg->listen_vcc,state_name[curr ? curr->state : ss_null]);
     msg->local.sas_addr.blli = NULL;
+    msg->remote.sas_addr.blli = NULL;
     bllis = msg2bllis(size);
-    this = &msg->remote.sas_addr.blli;
+    if (msg->type == as_bind || msg->type == as_listen)
+	this = &msg->local.sas_addr.blli;
+    else this = &msg->remote.sas_addr.blli;
     for (i = 0; i < bllis; i++) {
 	*this = &msg->blli[i];
 	this = &msg->blli[i].next;
