@@ -9,7 +9,7 @@
 #include <linux/atmsap.h>
 #include <linux/atmsvc.h>
 
-#include "timer.h"
+#include "atmd.h"
 
 
 typedef enum { /* formatting aligned with state_map and others */
@@ -53,6 +53,8 @@ extern const char *state_name[],*qs_name[],*as_name[];
 extern SOCKET *sockets;
 extern unsigned char q_buffer[];
 
+extern int net;
+ 
 
 #define msg2bllis(s) (((int) (s)-(int) sizeof(struct atmsvc_msg))/ \
   (int) sizeof(struct atm_blli)+1)
@@ -71,7 +73,9 @@ SOCKET *new_sock(unsigned long id);
 void free_sock(SOCKET *sock);
 void new_state(SOCKET *sock,STATE state);
 struct sockaddr_atmsvc *alloc_sap(const struct sockaddr_atmsvc *sap);
-void send_release(SOCKET *sock,unsigned char reason);
+void free_sap(struct sockaddr_atmsvc *sap);
+SOCKET *lookup_sap(struct sockaddr_atmsvc *sap);
+void send_release(SOCKET *sock,unsigned char reason,...);
 
 const char *mid2name(unsigned char mid);
 
