@@ -9,17 +9,17 @@
  */
 
 
+#include <stdarg.h>
 #include <stdio.h>
+#include <string.h>
 #include <assert.h>
 
+#include "atmd.h"
 #include "q2931.h"
 #include "qlib.h"
 
-#include "diag.h"
-#include "common.h"
 #include "proto.h"
 #include "io.h"
-#include "timer.h"
 #include "timeout.h"
 
 #define __KERNEL__ /* since that's what we effectively are */
@@ -60,23 +60,107 @@ static void send_release_complete(unsigned long call_ref,unsigned char cause)
     q_create(&dsc,q_buffer,MAX_Q_MSG);
     q_assign(&dsc,QF_msg_type,QMSG_REL_COMP);
     q_assign(&dsc,QF_call_ref,call_ref);
-    /* set cause */
+    /* set cause (only if !!cause) */
     size = q_close(&dsc);
     to_signaling(q_buffer,size);
 }
 
 
+static struct sockaddr_atmsvc *build_sap(void)
+{
+    struct sockaddr_atmsvc *sap;
+    struct atm_blli *blli;
+
+    sap = alloc_t(struct sockaddr_atmsvc);
+    memset(sap,0,sizeof(struct sockaddr_atmsvc));
+    if (q_present(&in_dsc,QF_aal_type))
+	if (q_fetch(&in_dsc,QF_aal_type) != 5)
+	    diag(COMPONENT,DIAG_ERROR,"AAL type %d requested",
+	      q_fetch(&in_dsc,QF_aal_type));
+    if (q_present(&in_dsc,QF_fw_pcr_01))
+	sap->sas_txtp.max_pcr = q_fetch(&in_dsc,QF_fw_pcr_01);
+    if (q_present(&in_dsc,QF_bw_pcr_01))
+	sap->sas_rxtp.max_pcr = q_fetch(&in_dsc,QF_bw_pcr_01);
+    if (q_present(&in_dsc,QF_fw_max_sdu))
+	sap->sas_txtp.max_sdu = q_fetch(&in_dsc,QF_fw_max_sdu);
+    if (q_present(&in_dsc,QF_bw_max_sdu))
+	sap->sas_rxtp.max_sdu = q_fetch(&in_dsc,QF_bw_max_sdu);
+    if (q_present(&in_dsc,QG_bhli)) {
+	sap->sas_addr.bhli.hl_type = q_fetch(&in_dsc,QF_hli_type);
+	switch (sap->sas_addr.bhli.hl_type) {
+	    case ATM_HL_ISO:
+		sap->sas_addr.bhli.hl_length = q_length(&in_dsc,QF_iso_hli);
+		q_read(&in_dsc,QF_iso_hli,sap->sas_addr.bhli.hl_info,
+		  sap->sas_addr.bhli.hl_length);
+		break;
+	    case ATM_HL_USER:
+		sap->sas_addr.bhli.hl_length = q_length(&in_dsc,QF_user_hli);
+		q_read(&in_dsc,QF_user_hli,sap->sas_addr.bhli.hl_info,
+		  sap->sas_addr.bhli.hl_length);
+		break;
+#ifdef UNI30
+	    case ATM_HL_HLP:
+		sap->sas_addr.bhli.hl_length = 4;
+		q_read(&in_dsc,QF_hlp,sap->sas_addr.bhli.hl_info,4);
+		break;
+#endif
+	    case ATM_HL_VENDOR:
+		sap->sas_addr.bhli.hl_length = 7;
+		q_read(&in_dsc,QF_hli_oui,sap->sas_addr.bhli.hl_info,3);
+		q_read(&in_dsc,QF_app_id,sap->sas_addr.bhli.hl_info+3,4);
+		break;
+	    default:
+		diag(COMPONENT,DIAG_FATAL,"unrecognized hl_type");
+	}
+    }
+
+#define GET(var,field) \
+   ({ if (q_present(&in_dsc,field)) blli->var = q_fetch(&in_dsc,field); })
+
+    if (q_present(&in_dsc,QG_blli)) {
+	sap->sas_addr.blli = blli = alloc_t(struct atm_blli);
+	if (q_present(&in_dsc,QF_uil2_proto)) {
+	    blli->l2_proto = q_fetch(&in_dsc,QF_uil2_proto);
+	    GET(l2.itu.mode,QF_l2_mode);
+	    GET(l2.itu.window,QF_window_size);
+	    GET(l2.user,QF_user_l2);
+	}
+	if (q_present(&in_dsc,QF_uil3_proto)) {
+	    blli->l3_proto = q_fetch(&in_dsc,QF_uil3_proto);
+	    GET(l3.itu.mode,QF_l3_mode);
+	    GET(l3.itu.def_size,QF_def_pck_size);
+	    GET(l3.itu.window,QF_pck_win_size);
+	    GET(l3.user,QF_user_l3);
+	    if (q_present(&in_dsc,QF_ipi_high)) {
+		blli->l3.tr9577.ipi = q_fetch(&in_dsc,QF_ipi_high) << 1;
+		if (blli->l3.tr9577.ipi != NLPID_IEEE802_1_SNAP)
+		    blli->l3.tr9577.ipi |= q_fetch(&in_dsc,QF_ipi_low);
+		else if (!q_present(&in_dsc,QF_oui)) blli->l3.tr9577.ipi |= 1;
+		    else {
+			q_read(&in_dsc,QF_oui,blli->l3.tr9577.snap,3);
+			q_read(&in_dsc,QF_pid,blli->l3.tr9577.snap+3,2);
+		    }
+	    }
+	}
+	blli->next = NULL;
+    }
+#undef GET
+    return sap;
+}
+
+
 static void setup_call(unsigned long call_ref)
 {
     SOCKET *sock,*this,**walk;
+    struct sockaddr_atmsvc *sap;
     unsigned char buf[ATM_ESA_LEN];
     int len,i;
 
-    for (sock = sockets; sock; sock = sock->next)
-	if (sock->state == ss_listening) break;
-    /* LISTENING */
+    sap = build_sap();
+    sock = lookup_sap(sap);
     if (!sock) {
-	send_release_complete(call_ref,0); /* @@@ dunno */
+	send_release_complete(call_ref,ATM_CV_INCOMP_DEST); /* @@@ dunno */
+	free_sap(sap);
 	return;
     }
     this = new_sock(0);
@@ -84,9 +168,9 @@ static void setup_call(unsigned long call_ref)
     this->q2931_state = qs_in_proc;
     this->call_ref = call_ref;
     send_call_proceeding(this);
-    /* this->local = alloc_sap(sock->local); @@@ - not yet */
+    this->local = alloc_sap(sock->local);
     printf("AAL type %ld\n",q_fetch(&in_dsc,QF_aal_type));
-    len = q_read(&in_dsc,QF_cdpn_nsap,(void *) &buf,sizeof(buf));
+    len = q_read(&in_dsc,QF_cdpn_esa,(void *) &buf,sizeof(buf));
     printf("Addr len %d:",len);
     for (i = 0; i < (len < 30 ? len : 30); i++) printf(" %02X",buf[i]);
     putchar('\n');
@@ -95,14 +179,14 @@ static void setup_call(unsigned long call_ref)
 	this->pvc.sap_addr.vci = q_fetch(&in_dsc,QF_vci);
 printf("VPI/VCI: %d/%d\n",this->pvc.sap_addr.vpi,this->pvc.sap_addr.vci);
     }
-    if (q_present(&in_dsc,QF_cgpn)) {
+    if (q_present(&in_dsc,QF_cgpn)) { /* improve @@@ */
 	this->remote = alloc_t(struct sockaddr_atmsvc);
-	this->remote->sas_addr.blli = NULL;
+	memset(this->remote,0,sizeof(struct sockaddr_atmsvc));
+	this->remote->sas_family = AF_ATMSVC;
 	i = q_read(&in_dsc,QF_cgpn,(void *) this->remote->sas_addr.prv,
 	  ATM_ESA_LEN);
 	printf("addr_len = %d\n",i);
     }
-    /* extract remote address */
     send_kernel(0,sock->id,as_indicate,0,this);
     for (walk = &sock->listen; *walk; walk = &(*walk)->listen);
     *walk = this;
@@ -111,8 +195,9 @@ printf("VPI/VCI: %d/%d\n",this->pvc.sap_addr.vpi,this->pvc.sap_addr.vci);
 }
 
 
-static void send_status(SOCKET *sock)
+static void send_status(SOCKET *sock,unsigned char cause,...)
 {
+    va_list ap;
     Q_DSC dsc;
     int size;
 
@@ -120,6 +205,16 @@ static void send_status(SOCKET *sock)
     q_assign(&dsc,QF_msg_type,QMSG_STATUS);
     q_assign(&dsc,QF_call_ref,sock->call_ref);
     q_assign(&dsc,QF_call_state,(int) sock->q2931_state);
+    q_assign(&dsc,QF_cause,cause);
+    va_start(ap,cause);
+    switch (cause) {
+	case ATM_CV_UNKNOWN_MSG_TYPE:
+	case ATM_CV_INCOMP_MSG:
+	    q_assign(&dsc,QF_bad_msg_type,va_arg(ap,unsigned char));
+	    break;
+	default:
+    }
+    va_end(ap);
     size = q_close(&dsc);
     to_signaling(q_buffer,size);
 }
@@ -143,11 +238,21 @@ static void q2931_call(SOCKET *sock,unsigned char mid)
     int error;
 
     switch (mid) {
-	case QMSG_STATUS:
-	    /* uh */
+	case QMSG_STATUS: /* 5.5.6.12 */
+	    {
+		Q2931_STATE state;
+
+		if (sock->q2931_state == qs_rel_req || sock->q2931_state ==
+		  qs_rel_ind) return;
+		state = q_fetch(&in_dsc,QF_call_state);
+		if (state == qs_null) break; /* clear call */
+		if (state != sock->q2931_state)
+		    diag(COMPONENT,DIAG_WARN,"STATUS %s received in state %s",
+		      qs_name[state],qs_name[sock->q2931_state]);
+	    }
 	    return;
 	case QMSG_STATUS_ENQ:
-	    send_status(sock);
+	    send_status(sock,ATM_CV_RESP_STAT_ENQ);
 	    return;
 	default:
 	    ;
@@ -155,13 +260,13 @@ static void q2931_call(SOCKET *sock,unsigned char mid)
     switch (mid) {
 	case QMSG_CALL_PROC: /* CONNECTING, WAIT_REL, REL_REQ */
 	    if (sock->state == ss_wait_rel || sock->state == ss_rel_req) {
-		send_status(sock);
+		send_status(sock,ATM_CV_INCOMP_MSG,QMSG_CALL_PROC);
 		return;
 	    }
 	    if (sock->state != ss_connecting) break;
 	    /* check for 2nd CALL_PROC @@@ */
 	    STOP_TIMER(sock);
-	    if (q_present(&in_dsc,QG_conn_id) >= 0) {
+	    if (q_present(&in_dsc,QG_conn_id)) {
 		sock->pvc.sap_addr.vpi = q_fetch(&in_dsc,QF_vpi);
 		sock->pvc.sap_addr.vci = q_fetch(&in_dsc,QF_vci);
 printf("VPI/VCI: %d/%d\n",sock->pvc.sap_addr.vpi,sock->pvc.sap_addr.vci);
@@ -171,12 +276,12 @@ printf("VPI/VCI: %d/%d\n",sock->pvc.sap_addr.vpi,sock->pvc.sap_addr.vci);
 	    return;
 	case QMSG_CONNECT: /* CONNECTING, REL_REQ */
 	    if (sock->state == ss_rel_req) {
-		send_status(sock);
+		send_status(sock,ATM_CV_INCOMP_MSG,QMSG_CONNECT);
 		return;
 	    }
 	    if (sock->state != ss_connecting) break;
 	    STOP_TIMER(sock);
-	    if (q_present(&in_dsc,QG_conn_id) >= 0) {
+	    if (q_present(&in_dsc,QG_conn_id)) {
 		sock->pvc.sap_addr.vpi = q_fetch(&in_dsc,QF_vpi);
 		sock->pvc.sap_addr.vci = q_fetch(&in_dsc,QF_vci);
 printf("VPI/VCI: %d/%d\n",sock->pvc.sap_addr.vpi,sock->pvc.sap_addr.vci);
@@ -187,7 +292,7 @@ printf("VPI/VCI: %d/%d\n",sock->pvc.sap_addr.vpi,sock->pvc.sap_addr.vci);
 	    /* more problems */
 	    if (error) {
 		set_error(sock,error);
-		send_release(sock,0); /* @@@ */
+		send_release(sock,0); /* @@@ cause follows reason ??? */
 		START_TIMER(sock,T308_1);
 		new_state(sock,ss_rel_req);
 		return;
@@ -204,7 +309,7 @@ printf("VPI/VCI: %d/%d\n",sock->pvc.sap_addr.vpi,sock->pvc.sap_addr.vci);
 	    diag(COMPONENT,DIAG_DEBUG,"CA vpi.vci=%d.%d",
 	      sock->pvc.sap_addr.vpi,sock->pvc.sap_addr.vci);
 	    if (sock->state == ss_wait_rel || sock->state == ss_rel_req) {
-		send_status(sock);
+		send_status(sock,ATM_CV_INCOMP_MSG,QMSG_CONN_ACK);
 		return;
 	    }
 	    if (sock->state != ss_accepting) break;
@@ -222,7 +327,7 @@ printf("VPI/VCI: %d/%d\n",sock->pvc.sap_addr.vpi,sock->pvc.sap_addr.vci);
 		    /* fall through */
 		case ss_accepting:
 		    set_error(sock,-ECONNRESET); /* ERESTARTSYS ? */
-		    send_release_complete(sock->call_ref,0); /* @@@ */
+		    send_release_complete(sock->call_ref,0);
 		    /* fall through */
 		case ss_rel_req:
 		    send_close(sock);
@@ -243,13 +348,15 @@ printf("VPI/VCI: %d/%d\n",sock->pvc.sap_addr.vpi,sock->pvc.sap_addr.vci);
 		    new_state(sock,ss_zombie);
 		    /* fall through */
 		case ss_rel_ind:
-		    /* send_release_complete(sock->call_ref,0); / * @@@ */
+		    /* send_release_complete(sock->call_ref,0); */
 		    return;
 		default:
-		    send_release_complete(sock->call_ref,0); /* @@@ */
+		    send_release_complete(sock->call_ref,0); /* @@@ should
+			be ATM_CV_INCOMP_MSG */
 		    break;
 	    }
 	    break;
+	case QMSG_STATUS: /* fall through when clearing */
 	case QMSG_REL_COMP: /* basically any state (except LISTENING,
 				  ZOMBIE, REL_IND) */
 	    switch (sock->state) {
@@ -282,13 +389,14 @@ printf("VPI/VCI: %d/%d\n",sock->pvc.sap_addr.vpi,sock->pvc.sap_addr.vci);
 	    }
 	default:
 	    diag(COMPONENT,DIAG_WARN,"Bad Q.2931 message %d",mid);
+	    send_status(sock,ATM_CV_UNKNOWN_MSG_TYPE,mid);
 	    return;
     }
     diag(COMPONENT,DIAG_WARN,
       "Q.2931 message %s is incompatible with state %s/%s (%d?%d)",
       mid2name(mid),state_name[sock->state],qs_name[sock->q2931_state],
       (int) sock->state,(int) sock->q2931_state);
-    send_status(sock);
+    send_status(sock,ATM_CV_INCOMP_MSG,mid);
 }
 
 
@@ -315,9 +423,11 @@ void to_q2931(void *msg,int size)
 	return;
     }
     if (!curr || curr->q2931_state == qs_null) {
-	if (mid != QMSG_REL_COMP && mid != QMSG_STATUS)
-	    send_release_complete(call_ref,81);
-		/* #81, "invalid call ref value" */
+	if (mid == QMSG_REL_COMP)
+	    send_release_complete(call_ref,ATM_CV_INV_CR);
+	else if (mid != QMSG_STATUS)
+		if (q_fetch(&in_dsc,QF_call_state) != (int) qs_null)
+		    send_release_complete(call_ref,ATM_CV_INCOMP_MSG);
 	return;
     }
     q2931_call(curr,mid);
