@@ -9,10 +9,9 @@
 #include <stdlib.h>
 #include <string.h>
 #include <netinet/in.h> /* for htonl, ntohl */
-#include "diag.h"
-#include "common.h"
-#include "timer.h"
-#include "queue.h"
+
+#include "atmd.h"
+
 #include "sscop.h"
 
 
@@ -158,10 +157,13 @@ static void maa_error(SSCOP_DSC *dsc,char code,int count)
     const char *const *walk;
 
     if (dsc->ops->maa_error)
-	if (!dsc->ops->maa_error(dsc,dsc->user,code,count)) return;
+	if (!dsc->ops->maa_error(dsc->user,code,count)) return;
     for (walk = msgs; **walk; walk++)
 	if (**walk == code) break;
-    diag(COMPONENT,DIAG_WARN,"layer management - error %c,%d \"%s\"",code,
+    if (code != 'V')
+	diag(COMPONENT,DIAG_WARN,"layer management - error %c \"%s\"",code,
+	  (*walk)+2);
+    else diag(COMPONENT,DIAG_WARN,"layer management - error %c,%d \"%s\"",code,
       count,(*walk)+2);
 }
 
@@ -456,7 +458,7 @@ static BUFFER *build_pdu(SSCOP_DSC *dsc,unsigned char type,void *data,
 	    *trailer++ = SSCOP_TRAIL(type,pad,dsc->vt_s);
 	    break;
 	case SSCOP_STAT:
-	    *trailer++ = SSCOP_TRAIL(0,0,dsc->vt_ps);
+	    *trailer++ = SSCOP_TRAIL(0,0,dsc->vr_ps);
 	    /* fall through */
 	case SSCOP_USTAT:
 	    *trailer++ = SSCOP_TRAIL(0,0,dsc->vr_mr);
@@ -479,7 +481,7 @@ static void dump_pdu(void *data,int len)
 {
     int i;
 
-    if (!dump) return;
+    if (!debug) return;
     for (i = 0; i < len; i++) {
 	if (!(i & 15)) printf("%s%04X:",i ? "\n" : "",i);
 	printf(" %02X",((unsigned char *) data)[i]);
@@ -501,7 +503,7 @@ static void emit_pdu(SSCOP_DSC *dsc,BUFFER *buf)
 	PRINT_PDU(dsc,"SEND",buf->data);
     }
     if (dsc->ops->cpcs_send)
-	dsc->ops->cpcs_send(dsc,dsc->user,buf->data,buf->length);
+	dsc->ops->cpcs_send(dsc->user,buf->data,buf->length);
     switch (SSCOP_TYPE(*(unsigned long *) ((char *) buf->data+buf->length-4))) {
 	case SSCOP_BGN:
 	    last = &dsc->last_bgn;
@@ -646,8 +648,7 @@ static void deliver_data(SSCOP_DSC *dsc) /* 20.49 */
     if (!dsc->clear_buffers) {
 	while ((buf = queue_get(&dsc->rx_buf))) {
 	    if (dsc->ops->data_ind)
-		dsc->ops->data_ind(dsc,dsc->user,buf->data,buf->length,
-		  buf->key);
+		dsc->ops->data_ind(dsc->user,buf->data,buf->length,buf->key);
 	    buffer_discard(buf);
 	}
     }
@@ -722,8 +723,7 @@ static void sscop_cc_exp(void *user)
 	    maa_error(dsc,'O',0);
 	    send_pdu(dsc,SSCOP_END,NULL,1);
 	    next_state(dsc,sscop_idle);
-	    if (dsc->ops->rel_ind)
-		dsc->ops->rel_ind(dsc,dsc->user,NULL,0,0);
+	    if (dsc->ops->rel_ind) dsc->ops->rel_ind(dsc->user,NULL,0,0);
 	    return;
 	case sscop_outdisc: /* 20.15 */
 	    if (dsc->vt_cc < dsc->cf_max_cc) {
@@ -734,7 +734,7 @@ static void sscop_cc_exp(void *user)
 	    }
 	    maa_error(dsc,'O',0);
 	    next_state(dsc,sscop_idle);
-	    if (dsc->ops->rel_conf) dsc->ops->rel_conf(dsc,dsc->user);
+	    if (dsc->ops->rel_conf) dsc->ops->rel_conf(dsc->user);
 	    return;
 	case sscop_outres: /* 20.18 */
 	    if (dsc->vt_cc < dsc->cf_max_cc) {
@@ -746,8 +746,7 @@ static void sscop_cc_exp(void *user)
 	    maa_error(dsc,'O',0);
 	    send_pdu(dsc,SSCOP_END,NULL,1);
 	    next_state(dsc,sscop_idle);
-	    if (dsc->ops->rel_ind)
-		dsc->ops->rel_ind(dsc,dsc->user,NULL,0,0);
+	    if (dsc->ops->rel_ind) dsc->ops->rel_ind(dsc->user,NULL,0,0);
 	    return;
 	case sscop_outrec: /* 20.24 */
 	    if (dsc->vt_cc < dsc->cf_max_cc) {
@@ -760,8 +759,7 @@ static void sscop_cc_exp(void *user)
 	    send_pdu(dsc,SSCOP_END,NULL,1);
 	    clear_receive_buffer(dsc);
 	    next_state(dsc,sscop_idle);
-	    if (dsc->ops->rel_ind)
-		dsc->ops->rel_ind(dsc,dsc->user,NULL,0,0);
+	    if (dsc->ops->rel_ind) dsc->ops->rel_ind(dsc->user,NULL,0,0);
 	    return;
 	default:
 	    break;
@@ -834,8 +832,7 @@ static void sscop_noresp_exp(void *user)
     send_pdu(dsc,SSCOP_END,NULL,1);
     prepare_retrieval(dsc);
     next_state(dsc,sscop_idle);
-    if (dsc->ops->rel_ind)
-	dsc->ops->rel_ind(dsc,dsc->user,NULL,0,0);
+    if (dsc->ops->rel_ind) dsc->ops->rel_ind(dsc->user,NULL,0,0);
 }
 
 
@@ -935,8 +932,7 @@ static void data_sd(SSCOP_DSC *dsc,int s,void *msg,int length) /* 20.38-39 */
 	return;
     }
     if (s == dsc->vr_r) {
-	if (dsc->ops->data_ind)
-	    dsc->ops->data_ind(dsc,dsc->user,msg,length,s);
+	if (dsc->ops->data_ind) dsc->ops->data_ind(dsc->user,msg,length,s);
 	if (s == dsc->vr_h) {
 	    dsc->vr_r = dsc->vr_h = MOD24(s+1);
 	    return;
@@ -947,8 +943,7 @@ static void data_sd(SSCOP_DSC *dsc,int s,void *msg,int length) /* 20.38-39 */
 	    if (!buf) break;
 	    queue_remove(&dsc->rx_buf,buf);
 	    if (dsc->ops->data_ind)
-		dsc->ops->data_ind(dsc,dsc->user,buf->data,buf->length,
-		  buf->key);
+		dsc->ops->data_ind(dsc->user,buf->data,buf->length,buf->key);
 	    buffer_discard(buf);
 	}
 	return;
@@ -980,7 +975,7 @@ static void data_sd(SSCOP_DSC *dsc,int s,void *msg,int length) /* 20.38-39 */
  */
 
 
-static void data_poll(SSCOP_DSC *dsc,int ps,int s) /* 20.41-42 */
+static void data_poll(SSCOP_DSC *dsc,int s) /* 20.41-42 */
 {
     int curr,i;
 
@@ -1157,22 +1152,25 @@ static void data_stat(SSCOP_DSC *dsc,int ps,int mr,int r,unsigned long *list,
 /* --- Incoming PDUs ------------------------------------------------------- */
 
 
-static void handle_sscop_pdu(SSCOP_DSC *dsc,void *msg,int size)
+/*
+ * Returns 0 if the descriptor might conceivably be gone when returning,
+ * 1 otherwise.
+ */
+ 
+static int handle_sscop_pdu(SSCOP_DSC *dsc,void *msg,int size)
 {
     PDU_VARS;
 
     dump_pdu(msg,size);
-    if (DECOMPOSE_PDU(dsc,msg,size)) return;
+    if (DECOMPOSE_PDU(dsc,msg,size)) return 1;
     PRINT_PDU(dsc,"RECV",msg);
     switch (type) {
 	case SSCOP_UD:
-	    if (dsc->ops->unitdata)
-		dsc->ops->unitdata(dsc,dsc->user,msg,length);
-	    return;
+	    if (dsc->ops->unitdata) dsc->ops->unitdata(dsc->user,msg,length);
+	    return 1;
 	case SSCOP_MD:
-	    if (dsc->ops->maa_data)
-		dsc->ops->maa_data(dsc,dsc->user,msg,length);
-	    return;
+	    if (dsc->ops->maa_data) dsc->ops->maa_data(dsc->user,msg,length);
+	    return 1;
 	default:
 	    break;
     }
@@ -1181,22 +1179,22 @@ static void handle_sscop_pdu(SSCOP_DSC *dsc,void *msg,int size)
 	    switch (type) {
 		case SSCOP_BGREJ:
 		    bad_pdu(dsc,type);
-		    return;
+		    return 1;
 		case SSCOP_BGN:
 		    if (detect_retransmission(dsc,sq)) {
 			send_pdu(dsc,SSCOP_BGREJ,NULL,0); /* no SSCOP-UU */
-			return;
+			return 1;
 		    }
 		    dsc->vt_ms = mr;
 		    next_state(dsc,sscop_inconn);
 		    if (dsc->ops->estab_ind)
-			dsc->ops->estab_ind(dsc,dsc->user,msg,length);
-		    return;
+			dsc->ops->estab_ind(dsc->user,msg,length);
+		    return 1;
 		case SSCOP_ENDAK:
-		    return;
+		    return 1;
 		case SSCOP_END:
 		    send_pdu(dsc,SSCOP_ENDAK,NULL,0);
-		    return;
+		    return 1;
 		case SSCOP_ER:
 		case SSCOP_POLL:
 		case SSCOP_SD:
@@ -1208,7 +1206,7 @@ static void handle_sscop_pdu(SSCOP_DSC *dsc,void *msg,int size)
 		case SSCOP_RSAK:
 		    bad_pdu(dsc,type);
 		    send_pdu(dsc,SSCOP_END,NULL,1);
-		    return;
+		    return 1;
 		default:
 		    break;
 	    }
@@ -1225,7 +1223,7 @@ static void handle_sscop_pdu(SSCOP_DSC *dsc,void *msg,int size)
 		case SSCOP_ER:
 		case SSCOP_RSAK:
 		case SSCOP_RS:
-		    return; /* ignore PDU */
+		    return 1; /* ignore PDU */
 		case SSCOP_BGAK:
 		    STOP_TIMER(dsc->timer_cc);
 		    dsc->vt_ms = mr;
@@ -1233,16 +1231,16 @@ static void handle_sscop_pdu(SSCOP_DSC *dsc,void *msg,int size)
 		    set_data_transfer_timers(dsc);
 		    next_state(dsc,sscop_ready);
 		    if (dsc->ops->estab_conf)
-			dsc->ops->estab_conf(dsc,dsc->user,msg,length);
-		    return;
+			dsc->ops->estab_conf(dsc->user,msg,length);
+		    return 1;
 		case SSCOP_BGREJ:
 		    STOP_TIMER(dsc->timer_cc);
 		    next_state(dsc,sscop_idle);
 		    if (dsc->ops->rel_ind)
-			dsc->ops->rel_ind(dsc,dsc->user,msg,length,1);
-		    return;
+			dsc->ops->rel_ind(dsc->user,msg,length,1);
+		    return 0;
 		case SSCOP_BGN:
-		    if (detect_retransmission(dsc,sq)) return;
+		    if (detect_retransmission(dsc,sq)) return 1;
 		    STOP_TIMER(dsc->timer_cc);
 		    dsc->vt_ms = mr;
 		    initialize_vr_mr(dsc);
@@ -1251,8 +1249,8 @@ static void handle_sscop_pdu(SSCOP_DSC *dsc,void *msg,int size)
 		    set_data_transfer_timers(dsc);
 		    next_state(dsc,sscop_ready);
 		    if (dsc->ops->estab_conf)
-			dsc->ops->estab_conf(dsc,dsc->user,msg,length);
-		    return;
+			dsc->ops->estab_conf(dsc->user,msg,length);
+		    return 1;
 		default:
 		    break;
 	    }
@@ -1260,25 +1258,25 @@ static void handle_sscop_pdu(SSCOP_DSC *dsc,void *msg,int size)
 	case sscop_inconn: /* 20.11-13 */
 	    switch (type) {
 		case SSCOP_BGN:
-		    if (detect_retransmission(dsc,sq)) return;
+		    if (detect_retransmission(dsc,sq)) return 1;
 		    dsc->vt_ms = mr;
 		    if (dsc->ops->restart)
-			dsc->ops->restart(dsc,dsc->user,msg,length,1);
-		    return;
+			dsc->ops->restart(dsc->user,msg,length,1);
+		    return 1;
 		case SSCOP_ER:
 		case SSCOP_BGAK:
 		case SSCOP_ERAK:
 		case SSCOP_RSAK:
 		case SSCOP_RS:
 		    bad_pdu(dsc,type);
-		    return;
+		    return 1;
 		case SSCOP_ENDAK:
 		case SSCOP_BGREJ:
 		    bad_pdu(dsc,type);
 		    next_state(dsc,sscop_idle);
 		    if (dsc->ops->rel_ind)
-			dsc->ops->rel_ind(dsc,dsc->user,NULL,0,1);
-		    return;
+			dsc->ops->rel_ind(dsc->user,NULL,0,1);
+		    return 0;
 		case SSCOP_SD:
 		case SSCOP_USTAT:
 		case SSCOP_STAT:
@@ -1287,14 +1285,14 @@ static void handle_sscop_pdu(SSCOP_DSC *dsc,void *msg,int size)
 		    send_pdu(dsc,SSCOP_END,NULL,1);
 		    next_state(dsc,sscop_idle);
 		    if (dsc->ops->rel_ind)
-			dsc->ops->rel_ind(dsc,dsc->user,NULL,0,0);
-		    return;
+			dsc->ops->rel_ind(dsc->user,NULL,0,0);
+		    return 0;
 		case SSCOP_END:
 		    send_pdu(dsc,SSCOP_ENDAK,NULL,0);
 		    next_state(dsc,sscop_idle);
 		    if (dsc->ops->rel_ind)
-			dsc->ops->rel_ind(dsc,dsc->user,msg,length,!s);
-		    return;
+			dsc->ops->rel_ind(dsc->user,msg,length,!s);
+		    return 0;
 		default:
 		    break;
 	    }
@@ -1310,33 +1308,31 @@ static void handle_sscop_pdu(SSCOP_DSC *dsc,void *msg,int size)
 		case SSCOP_RS:
 		case SSCOP_RSAK:
 		case SSCOP_ER:
-		    return;
+		    return 1;
 		case SSCOP_END:
 		    STOP_TIMER(dsc->timer_cc);
 		    send_pdu(dsc,SSCOP_ENDAK,NULL,0);
 		    next_state(dsc,sscop_idle);
-		    if (dsc->ops->rel_conf)
-			dsc->ops->rel_conf(dsc,dsc->user);
-		    return;
+		    if (dsc->ops->rel_conf) dsc->ops->rel_conf(dsc->user);
+		    return 0;
 		case SSCOP_ENDAK:
 		case SSCOP_BGREJ:
 		    STOP_TIMER(dsc->timer_cc);
 		    next_state(dsc,sscop_idle);
-		    if (dsc->ops->rel_conf)
-			dsc->ops->rel_conf(dsc,dsc->user);
-		    return;
+		    if (dsc->ops->rel_conf) dsc->ops->rel_conf(dsc->user);
+		    return 0;
 		case SSCOP_BGN:
 		    if (detect_retransmission(dsc,sq)) {
 			send_pdu(dsc,SSCOP_BGAK,NULL,0);
 			resend_pdu(dsc,dsc->last_end);
-			return;
+			return 1;
 		    }
 		    STOP_TIMER(dsc->timer_cc);
 		    dsc->vt_ms = mr;
 		    next_state(dsc,sscop_inconn);
 		    if (dsc->ops->restart)
-			dsc->ops->restart(dsc,dsc->user,msg,length,0);
-		    return;
+			dsc->ops->restart(dsc->user,msg,length,0);
+		    return 1;
 	    }
 	case sscop_outres: /* 20.17-19 */
 	    switch (type) {
@@ -1347,36 +1343,36 @@ static void handle_sscop_pdu(SSCOP_DSC *dsc,void *msg,int size)
 		case SSCOP_BGAK:
 		case SSCOP_ERAK:
 		case SSCOP_SD:
-		    return; /* ignore */
+		    return 1; /* ignore */
 		case SSCOP_BGN:
 		    if (detect_retransmission(dsc,sq)) {
 			send_pdu(dsc,SSCOP_BGAK,NULL,0);
 			resend_pdu(dsc,dsc->last_rs);
-			return;
+			return 1;
 		    }
 		    STOP_TIMER(dsc->timer_cc);
 		    dsc->vt_ms = mr;
 		    next_state(dsc,sscop_inconn);
 		    if (dsc->ops->restart)
-			dsc->ops->restart(dsc,dsc->user,msg,length,1);
-		    return;
+			dsc->ops->restart(dsc->user,msg,length,1);
+		    return 1;
 		case SSCOP_ENDAK:
 		case SSCOP_BGREJ:
 		    STOP_TIMER(dsc->timer_cc);
 		    bad_pdu(dsc,type);
 		    next_state(dsc,sscop_idle);
 		    if (dsc->ops->rel_ind)
-			dsc->ops->rel_ind(dsc,dsc->user,NULL,0,0);
-		    return;
+			dsc->ops->rel_ind(dsc->user,NULL,0,0);
+		    return 0;
 		case SSCOP_END:
 		    STOP_TIMER(dsc->timer_cc);
 		    send_pdu(dsc,SSCOP_ENDAK,NULL,0);
 		    next_state(dsc,sscop_idle);
 		    if (dsc->ops->rel_ind)
-			dsc->ops->rel_ind(dsc,dsc->user,msg,length,!s);
-		    return;
+			dsc->ops->rel_ind(dsc->user,msg,length,!s);
+		    return 0;
 		case SSCOP_RS:
-		    if (detect_retransmission(dsc,sq)) return;
+		    if (detect_retransmission(dsc,sq)) return 1;
 		    STOP_TIMER(dsc->timer_cc);
 		    dsc->vt_ms = mr;
 		    initialize_vr_mr(dsc);
@@ -1384,18 +1380,16 @@ static void handle_sscop_pdu(SSCOP_DSC *dsc,void *msg,int size)
 		    initialize_state_variables(dsc);
 		    set_data_transfer_timers(dsc);
 		    next_state(dsc,sscop_ready);
-		    if (dsc->ops->res_conf)
-			dsc->ops->res_conf(dsc,dsc->user);
-		    return;
+		    if (dsc->ops->res_conf) dsc->ops->res_conf(dsc->user);
+		    return 1;
 		case SSCOP_RSAK:
 		    STOP_TIMER(dsc->timer_cc);
 		    dsc->vt_ms = mr;
 		    initialize_state_variables(dsc);
 		    set_data_transfer_timers(dsc);
 		    next_state(dsc,sscop_ready);
-		    if (dsc->ops->res_conf)
-			dsc->ops->res_conf(dsc,dsc->user);
-		    return;
+		    if (dsc->ops->res_conf) dsc->ops->res_conf(dsc->user);
+		    return 1;
 		default:
 		    break;
 	    }
@@ -1413,33 +1407,33 @@ static void handle_sscop_pdu(SSCOP_DSC *dsc,void *msg,int size)
 		    bad_pdu(dsc,type);
 		    next_state(dsc,sscop_idle);
 		    if (dsc->ops->rel_ind)
-			dsc->ops->rel_ind(dsc,dsc->user,NULL,0,0);
-		    return;
+			dsc->ops->rel_ind(dsc->user,NULL,0,0);
+		    return 0;
 		case SSCOP_END:
 		    send_pdu(dsc,SSCOP_ENDAK,NULL,0);
 		    next_state(dsc,sscop_idle);
 		    if (dsc->ops->rel_ind)
-			dsc->ops->rel_ind(dsc,dsc->user,msg,length,!s);
-		    return;
+			dsc->ops->rel_ind(dsc->user,msg,length,!s);
+		    return 0;
 		case SSCOP_ER:
 		case SSCOP_BGAK:
 		case SSCOP_ERAK:
 		case SSCOP_RSAK:
 		    bad_pdu(dsc,type);
-		    return;
+		    return 1;
 		case SSCOP_BGN:
 		    if (detect_retransmission(dsc,sq)) {
 			bad_pdu(dsc,type);
-			return;
+			return 1;
 		    }
 		    dsc->vt_ms = mr;
 		    next_state(dsc,sscop_inconn);
 		    if (dsc->ops->restart)
-			dsc->ops->restart(dsc,dsc->user,msg,length,1);
-		    return;
+			dsc->ops->restart(dsc->user,msg,length,1);
+		    return 1;
 		case SSCOP_RS:
 		    if (!detect_retransmission(dsc,sq)) bad_pdu(dsc,type);
-		    return;
+		    return 1;
 		default:
 		    break;
 	    }
@@ -1449,23 +1443,22 @@ static void handle_sscop_pdu(SSCOP_DSC *dsc,void *msg,int size)
 		case SSCOP_BGAK:
 		case SSCOP_RSAK:
 		    bad_pdu(dsc,type);
-		    return;
+		    return 1;
 		case SSCOP_ERAK:
 		    STOP_TIMER(dsc->timer_cc);
 		    dsc->vt_ms = mr;
 		    next_state(dsc,sscop_recresp);
 		    deliver_data(dsc);
-		    if (dsc->ops->rec_ind)
-			dsc->ops->rec_ind(dsc,dsc->user);
-		    return;
+		    if (dsc->ops->rec_ind) dsc->ops->rec_ind(dsc->user);
+		    return 1;
 		case SSCOP_END:
 		    STOP_TIMER(dsc->timer_cc);
 		    send_pdu(dsc,SSCOP_ENDAK,NULL,0);
 		    clear_receive_buffer(dsc);
 		    next_state(dsc,sscop_idle);
 		    if (dsc->ops->rel_ind)
-			dsc->ops->rel_ind(dsc,dsc->user,msg,length,!s);
-		    return;
+			dsc->ops->rel_ind(dsc->user,msg,length,!s);
+		    return 0;
 		case SSCOP_ENDAK:
 		case SSCOP_BGREJ:
 		    bad_pdu(dsc,type);
@@ -1473,29 +1466,29 @@ static void handle_sscop_pdu(SSCOP_DSC *dsc,void *msg,int size)
 		    clear_receive_buffer(dsc);
 		    next_state(dsc,sscop_idle);
 		    if (dsc->ops->rel_ind)
-			dsc->ops->rel_ind(dsc,dsc->user,NULL,0,0);
-		    return;
+			dsc->ops->rel_ind(dsc->user,NULL,0,0);
+		    return 0;
 		case SSCOP_STAT:
 		case SSCOP_USTAT:
 		case SSCOP_POLL:
 		case SSCOP_SD:
-		    return;
+		    return 1;
 		case SSCOP_BGN:
 		    if (detect_retransmission(dsc,sq)) {
 			bad_pdu(dsc,type);
-			return;
+			return 1;
 		    }
 		    STOP_TIMER(dsc->timer_cc);
 		    dsc->vt_ms = mr;
 		    clear_receive_buffer(dsc);
 		    next_state(dsc,sscop_inconn);
 		    if (dsc->ops->restart)
-			dsc->ops->restart(dsc,dsc->user,msg,length,1);
-		    return;
+			dsc->ops->restart(dsc->user,msg,length,1);
+		    return 1;
 		case SSCOP_ER:
 		    if (detect_retransmission(dsc,sq)) {
 			bad_pdu(dsc,type);
-			return;
+			return 1;
 		    }
 		    STOP_TIMER(dsc->timer_cc);
 		    dsc->vt_ms = mr;
@@ -1503,21 +1496,20 @@ static void handle_sscop_pdu(SSCOP_DSC *dsc,void *msg,int size)
 		    send_pdu(dsc,SSCOP_ERAK,NULL,0);
 		    deliver_data(dsc);
 		    next_state(dsc,sscop_recresp);
-		    if (dsc->ops->rec_ind)
-			dsc->ops->rec_ind(dsc,dsc->user);
-		    return;
+		    if (dsc->ops->rec_ind) dsc->ops->rec_ind(dsc->user);
+		    return 1;
 		case SSCOP_RS:
 		    if (detect_retransmission(dsc,sq)) {
 			bad_pdu(dsc,type);
-			return;
+			return 1;
 		    }
 		    STOP_TIMER(dsc->timer_cc);
 		    dsc->vt_ms = mr;
 		    clear_receive_buffer(dsc);
 		    next_state(dsc,sscop_inres);
 		    if (dsc->ops->res_ind)
-			dsc->ops->res_ind(dsc,dsc->user,msg,length);
-		    return;
+			dsc->ops->res_ind(dsc->user,msg,length);
+		    return 0;
 		default:
 		    break;
 	    }
@@ -1526,56 +1518,56 @@ static void handle_sscop_pdu(SSCOP_DSC *dsc,void *msg,int size)
 		case SSCOP_BGAK:
 		case SSCOP_RSAK:
 		    bad_pdu(dsc,type);
-		    return;
+		    return 1;
 		case SSCOP_ERAK:
 		case SSCOP_SD:
 		case SSCOP_POLL:
-		    return;
+		    return 1;
 		case SSCOP_END:
 		    send_pdu(dsc,SSCOP_ENDAK,NULL,0);
 		    next_state(dsc,sscop_idle);
 		    if (dsc->ops->rel_ind)
-			dsc->ops->rel_ind(dsc,dsc->user,msg,length,!s);
-		    return;
+			dsc->ops->rel_ind(dsc->user,msg,length,!s);
+		    return 0;
 		case SSCOP_ENDAK:
 		case SSCOP_BGREJ:
 		    bad_pdu(dsc,type);
 		    next_state(dsc,sscop_idle);
 		    if (dsc->ops->rel_ind)
-			dsc->ops->rel_ind(dsc,dsc->user,NULL,0,0);
-		    return;
+			dsc->ops->rel_ind(dsc->user,NULL,0,0);
+		    return 0;
 		case SSCOP_RS:
 		    if (detect_retransmission(dsc,sq)) {
 			bad_pdu(dsc,type);
-			return;
+			return 1;
 		    }
 		    dsc->vt_ms = mr;
 		    next_state(dsc,sscop_inres);
 		    if (dsc->ops->res_ind)
-			dsc->ops->res_ind(dsc,dsc->user,msg,length);
-		    return;
+			dsc->ops->res_ind(dsc->user,msg,length);
+		    return 1;
 		case SSCOP_ER:
 		    if (!detect_retransmission(dsc,sq)) bad_pdu(dsc,type);
 		    else send_pdu(dsc,SSCOP_ERAK,NULL,0);
-		    return;
+		    return 1;
 		case SSCOP_BGN:
 		    if (detect_retransmission(dsc,sq)) {
 			bad_pdu(dsc,type);
-			return;
+			return 1;
 		    }
 		    dsc->vt_ms = mr;
 		    next_state(dsc,sscop_inconn);
 		    if (dsc->ops->restart)
-			dsc->ops->restart(dsc,dsc->user,msg,length,1);
-		    return;
+			dsc->ops->restart(dsc->user,msg,length,1);
+		    return 1;
 		case SSCOP_STAT:
 		case SSCOP_USTAT:
 		    bad_pdu(dsc,type);
 		    send_pdu(dsc,SSCOP_END,NULL,1);
 		    next_state(dsc,sscop_idle);
 		    if (dsc->ops->rel_ind)
-			dsc->ops->rel_ind(dsc,dsc->user,NULL,0,0);
-		    return;
+			dsc->ops->rel_ind(dsc->user,NULL,0,0);
+		    return 0;
 		default:
 		    break;
 	    }
@@ -1585,15 +1577,15 @@ static void handle_sscop_pdu(SSCOP_DSC *dsc,void *msg,int size)
 		    send_pdu(dsc,SSCOP_ENDAK,NULL,0);
 		    next_state(dsc,sscop_idle);
 		    if (dsc->ops->rel_ind)
-			dsc->ops->rel_ind(dsc,dsc->user,msg,length,!s);
-		    return;
+			dsc->ops->rel_ind(dsc->user,msg,length,!s);
+		    return 0;
 		case SSCOP_ENDAK:
 		case SSCOP_BGREJ:
 		    bad_pdu(dsc,type);
 		    next_state(dsc,sscop_idle);
 		    if (dsc->ops->rel_ind)
-			dsc->ops->rel_ind(dsc,dsc->user,NULL,0,0);
-		    return;
+			dsc->ops->rel_ind(dsc->user,NULL,0,0);
+		    return 0;
 		case SSCOP_USTAT:
 		case SSCOP_STAT:
 		case SSCOP_POLL:
@@ -1602,36 +1594,36 @@ static void handle_sscop_pdu(SSCOP_DSC *dsc,void *msg,int size)
 		    send_pdu(dsc,SSCOP_END,NULL,1);
 		    next_state(dsc,sscop_idle);
 		    if (dsc->ops->rel_ind)
-			dsc->ops->rel_ind(dsc,dsc->user,NULL,0,0);
-		    return;
+			dsc->ops->rel_ind(dsc->user,NULL,0,0);
+		    return 0;
 		case SSCOP_RSAK:
 		case SSCOP_BGAK:
 		case SSCOP_ERAK:
 		    bad_pdu(dsc,type);
-		    return;
+		    return 1;
 		case SSCOP_RS:
 		    if (detect_retransmission(dsc,sq)) {
 			bad_pdu(dsc,type);
-			return;
+			return 1;
 		    }
 		    dsc->vt_ms = mr;
 		    next_state(dsc,sscop_inres);
 		    if (dsc->ops->res_ind)
-			dsc->ops->res_ind(dsc,dsc->user,msg,length);
-		    return;
+			dsc->ops->res_ind(dsc->user,msg,length);
+		    return 1;
 		case SSCOP_ER:
 		    if (!detect_retransmission(dsc,sq)) bad_pdu(dsc,type);
-		    return;
+		    return 1;
 		case SSCOP_BGN:
 		    if (detect_retransmission(dsc,sq)) {
 			bad_pdu(dsc,type);
-			return;
+			return 1;
 		    }
 		    dsc->vt_ms = mr;
 		    next_state(dsc,sscop_inconn);
 		    if (dsc->ops->restart)
-			dsc->ops->restart(dsc,dsc->user,msg,length,1);
-		    return;
+			dsc->ops->restart(dsc->user,msg,length,1);
+		    return 1;
 		default:
 		    break;
 	    }
@@ -1640,34 +1632,33 @@ static void handle_sscop_pdu(SSCOP_DSC *dsc,void *msg,int size)
 		case SSCOP_BGAK:
 		case SSCOP_ERAK:
 		case SSCOP_RSAK:
-		    return;
+		    return 1;
 		case SSCOP_ER:
 		    if (detect_retransmission(dsc,sq)) {
 			START_TIMER(noresp);
 			send_pdu(dsc,SSCOP_ERAK,NULL,0);
-			return;
+			return 1;
 		    }
 		    reset_data_transfer_timers(dsc);
 		    dsc->vt_ms = mr;
 		    prepare_recovery(dsc);
 		    next_state(dsc,sscop_inrec);
 		    deliver_data(dsc);
-		    if (dsc->ops->rec_ind)
-			dsc->ops->rec_ind(dsc,dsc->user);
-		    return;
+		    if (dsc->ops->rec_ind) dsc->ops->rec_ind(dsc->user);
+		    return 1;
 		case SSCOP_BGN:
 		    if (detect_retransmission(dsc,sq)) {
 			START_TIMER(noresp);
 			send_pdu(dsc,SSCOP_BGAK,NULL,0);
-			return;
+			return 1;
 		    }
 		    reset_data_transfer_timers(dsc);
 		    dsc->vt_ms = mr;
 		    prepare_retrieval(dsc);
 		    next_state(dsc,sscop_inconn);
 		    if (dsc->ops->restart)
-			dsc->ops->restart(dsc,dsc->user,msg,length,1);
-		    return;
+			dsc->ops->restart(dsc->user,msg,length,1);
+		    return 1;
 		case SSCOP_ENDAK:
 		case SSCOP_BGREJ:
 		    reset_data_transfer_timers(dsc);
@@ -1675,53 +1666,55 @@ static void handle_sscop_pdu(SSCOP_DSC *dsc,void *msg,int size)
 		    prepare_retrieval(dsc);
 		    next_state(dsc,sscop_idle);
 		    if (dsc->ops->rel_ind)
-			dsc->ops->rel_ind(dsc,dsc->user,NULL,0,0);
-		    return;
+			dsc->ops->rel_ind(dsc->user,NULL,0,0);
+		    return 0;
 		case SSCOP_RS:
 		    if (detect_retransmission(dsc,sq)) {
 			START_TIMER(noresp);
 			send_pdu(dsc,SSCOP_RSAK,NULL,0);
-			return;
+			return 1;
 		    }
 		    reset_data_transfer_timers(dsc);
 		    dsc->vt_ms = mr;
 		    prepare_retrieval(dsc);
 		    next_state(dsc,sscop_inres);
 		    if (dsc->ops->res_ind)
-			dsc->ops->res_ind(dsc,dsc->user,msg,length);
-		    return;
+			dsc->ops->res_ind(dsc->user,msg,length);
+		    return 1;
 		case SSCOP_END:
 		    reset_data_transfer_timers(dsc);
 		    send_pdu(dsc,SSCOP_ENDAK,NULL,0);
 		    prepare_retrieval(dsc);
 		    next_state(dsc,sscop_idle);
 		    if (dsc->ops->rel_ind)
-			dsc->ops->rel_ind(dsc,dsc->user,msg,length,!s);
-		    return;
+			dsc->ops->rel_ind(dsc->user,msg,length,!s);
+		    return 0;
 		case SSCOP_SD:
 		    data_sd(dsc,s,msg,length);
-		    return;
+		    return 1;
 		case SSCOP_POLL:
-		    data_poll(dsc,ps,s);
-		    return;
+		    dsc->vr_ps = ps; /* store */
+		    data_poll(dsc,s);
+		    return 1;
 		case SSCOP_USTAT:
 		    data_ustat(dsc,mr,r,SSCOP_N(((unsigned long *) msg)[0]),
 		      SSCOP_N(((unsigned long *) msg)[1]));
-		    return;
+		    return 1;
 		case SSCOP_STAT:
 		    data_stat(dsc,ps,mr,r,msg,length/4);
-		    return;
+		    return 1;
 		default:
 		    break;
 	    }
     }
+    return 1;
 }
 
 
 void sscop_pdu(SSCOP_DSC *dsc,void *msg,int size)
 {
-    handle_sscop_pdu(dsc,msg,size);
-    if (dsc->state == sscop_ready) try_to_send(dsc);
+    if (handle_sscop_pdu(dsc,msg,size))
+	if (dsc->state == sscop_ready) try_to_send(dsc);
 }
 
 
@@ -1746,16 +1739,16 @@ void sscop_retrieve(SSCOP_DSC *dsc,int rn)
 	    if (buf) {
 		queue_remove(&dsc->tx_buf,buf);
 		if (dsc->ops->retr_ind)
-		    dsc->ops->retr_ind(dsc,dsc->user,buf->data,buf->length);
+		    dsc->ops->retr_ind(dsc->user,buf->data,buf->length);
 		buffer_discard(buf);
 	    }
 	}
     while ((buf = queue_get(&dsc->tx_q))) {
 	if (dsc->ops->retr_ind)
-	    dsc->ops->retr_ind(dsc,dsc->user,buf->data,buf->length);
+	    dsc->ops->retr_ind(dsc->user,buf->data,buf->length);
 	buffer_discard(buf);
     }
-    if (dsc->ops->retr_comp) dsc->ops->retr_comp(dsc,dsc->user);
+    if (dsc->ops->retr_comp) dsc->ops->retr_comp(dsc->user);
 }
 
 
